# -*- coding: utf-8 -*-

import salt.ext.six as six
import salt.state
import salt.utils.args

_keywords = salt.state.STATE_REQUISITE_KEYWORDS | \
            salt.state.STATE_REQUISITE_IN_KEYWORDS | \
            salt.state.STATE_RUNTIME_KEYWORDS | \
            {'name'}


def test_match_args(self, state):
    if state['state'] not in self.mods:
        self._logger('State module does not exist: %s', state['state'])
        return
    if not hasattr(self.mods[state['state']], state['fun']):
        self._logger('State function does not exist: %s.%s', state['state'], state['fun'])
        return
    args = salt.utils.args.get_function_argspec(getattr(self.mods[state['state']], state['fun']))
    if args.keywords is not None:
        return
    for arg in six.iterkeys(state):
        if arg not in args and arg not in _keywords:
            self._logger(
                'Arg not accepted by state %s.%s: %s',
                state['state'], state['fun'], arg,
            )



# -*- coding: utf-8 -*-


def test_bad_yaml_dict(self, state):
    this = self.rendered[state['__id__']]['.'.join([state['state'], state['fun']])]
    for dict_ in this:
        if len(dict_) > 1 and None in dict_.values():
            for key in dict_:
                self._logger('Dictionary values not indented enough: %s is None', key)
                return



# -*- coding: utf-8 -*-
import os

import salint
import salint.parser


def _ensure_abs(path):
    if not os.path.isabs(path):
        return os.path.join(os.getcwd(), path)
    return path


def main():
    args = salint.parser.parse_args()
    paths = [_ensure_abs(path) for path in args.paths or [os.getcwd()]]
    opts = {
        'fileserver_backend': ['roots'],
        'extension_modules': './.lintcache/extmods',
        'file_client': 'local',
        'cachedir': './.lintcache/',
        'file_roots': {
            'base': paths,
        },
        'environment': 'base',
        'state_top': 'salt://top.sls',
        'state_auto_order': True,
        'fileserver_followsymlinks': True,
        'fileserver_ignoresymlinks': False,
        'id': 'local',
        'saltenv': 'base',
        'pillar_cache': False,
        'pillar_roots': {},
        'renderer_blacklist': [],
        'renderer_whitelist': [],
        'hash_type': 'sha256',
        'file_buffer_size': 262144,
        'renderer': 'yaml_jinja',
        'log_level': 'critical',
        'file_ignore_regex': [],
        'file_ignore_glob': [],
        'cython_enable': False,
    }
    lint = salint.Linter(opts)
    lint.lint_all()



# -*- coding: utf-8 -*-

import salt.state
import salt.ext.six as six


def test_requisites_match(self, state):

    def _sls_test(sls):
        return key == sls['state'] and (value == sls['__id__'] or value == sls['name'])

    for req_word in salt.state.STATE_REQUISITE_KEYWORDS | salt.state.STATE_REQUISITE_IN_KEYWORDS:
        for req in state.get(req_word, []):
            if isinstance(req, six.string_types):
                if not any(filter(lambda sls: req == sls['__sls__'], self.lowstate)):
                    self._logger('Unable to find requisite: %s', req)
            elif isinstance(req, dict):
                key, value = next(six.iteritems(req))
                if not any(filter(_sls_test, self.lowstate)):
                    self._logger('Unable to find requisite: %s', dict(req))
            else:
                self._logger('Unrecognized requisite type: %s', type(req))



# -*- coding: utf-8 -*-
import argparse


def parse_args():
    parser = argparse.ArgumentParser(description='Lint Salt States.')
    parser.add_argument('paths', metavar='PATH', type=str, nargs='*',
                        help="paths to search for sls files in")
    return parser.parse_args()



# -*- coding: utf-8 -*-
'''
Setup File for Salint
'''
from setuptools import setup
setup(
    name='salint',
    version='0.0.2',
    description='Linter for SaltStack States',
    url='https://github.com/gtmanfred/salint',
    author='Daniel Wallace',
    author_email='daniel@saltstack.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=['salint', 'salint.rules'],
    entry_points={  # Optional
        'console_scripts': [
            'salint=salint.script:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/gtmanfred/salint/issues',
        'Source': 'https://github.com/gtmanfred/salint/',
    },
)



# -*- coding: utf-8 -*-
import inspect
import logging
import os
import pkg_resources
import sys

import salt.loader
import salt.state
import salt.ext.six as six

import salint.rules

if six.PY2:
    import imp

    def load_module(fullname, path):
        fp, pathname, description = imp.find_module(fullname, [path])
        return imp.load_module(fullname, fp, pathname, description)
else:
    import importlib

    def load_module(fullname, path):
        return importlib.machinery.PathFinder.find_module(fullname, [path]).load_module()

log_format = '%(statefile)s | %(stateid)s | %(statename)s | %(message)s'
date_format = '%H:%M:%S'
formatter = logging.Formatter(log_format, datefmt=date_format)

log = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stderr)
handler.setLevel('warning'.upper())
handler.setFormatter(formatter)
logging.root.addHandler(handler)


class Linter(object):
    def __init__(self, opts=None):
        self.opts = opts or {}
        self._load_modules()
        self._load_renderers()

    def _load_modules(self):
        dict(salt.loader.LazyLoader(
            salt.loader._module_dirs(self.opts, 'states', 'states'),
            self.opts,
            tag='lint',
            virtual_enable=False,
        ))
        self.mods = {}
        for module in filter(lambda mod: mod.startswith('salt.loaded.int.lint'), sys.modules):
            mod = sys.modules[module]
            self.mods[getattr(mod, '__virtualname__', module.split('.')[-1])] = mod
            for alias in getattr(mod, '__virtual_aliases__', []):
                self.mods[alias] = mod

    def _load_renderers(self):
        self.minion_mods = salt.loader.minion_mods(self.opts)
        self.renderers = salt.loader.render(
            self.opts,
            self.minion_mods,
        )

    def get_low(self, mods):
        st_ = salt.state.HighState(self.opts)

        if isinstance(mods, six.string_types):
            mods = mods.split(',')
        st_.push_active()
        try:
            high_, errors = st_.render_highstate({'base': mods})
        finally:
            st_.pop_active()
        errors = st_.state.verify_high(high_)
        if errors:
            return errors
        return st_.state.compile_high_data(high_)

    def _logger(self, message, *args, **kwargs):
        level = kwargs.pop('level', 'warning')
        getattr(log, level)(message, *args, extra=self.info)

    def _get_members(self, mod):
        def _get_test_functions(fun):
            return inspect.isfunction(fun) and fun.__name__.startswith('test_')
        return inspect.getmembers(mod, predicate=_get_test_functions)

    @property
    def tests(self):
        for parent, _, rulefiles in os.walk(os.path.dirname(salint.rules.__file__)):
            for rulefile in rulefiles:
                if not rulefile.endswith('.py') or rulefile == '__init__.py':
                    continue
                for ret in self._get_members(load_module(rulefile[:-3], parent)):
                    yield ret[1]
        for rulefile in pkg_resources.iter_entry_points('salint.rules', name=None):
            for ret in self._get_members(rulefile.load()):
                yield ret[1]

    def render_state(self, path, parent=None):
        if parent:
            path = os.path.join(parent, path)
        if not os.path.isabs(path):
            path = '/' + path
        return salt.template.compile_template(
            path,
            self.renderers,
            'jinja|yaml',
            self.opts['renderer_blacklist'],
            self.opts['renderer_whitelist'],
        )

    def lint_sls(self, sls, path=None):
        self.lowstate = self.get_low(sls)
        self.rendered = self.render_state(os.path.join(*sls.split('.')) + '.sls', parent=path)

        for state in self.lowstate:
            if sls != state['__sls__']:
                continue
            self.info = {
                'stateid': state['__id__'],
                'statename': state['name'],
                'statefile': state['__sls__'] + '.sls',
            }
            for test in self.tests:
                test(self, state)

    def lint_all(self):
        for dir_ in self.opts['file_roots']['base']:
            dir_ = dir_.rstrip()
            for parent, _, statefiles in os.walk(dir_):
                if parent.startswith('.') and parent not in ('.', '..'):
                    continue
                for sls in statefiles:
                    if sls.endswith('.sls'):
                        if parent == dir_:
                            self.lint_sls(sls[:-4], path=dir_)
                        else:
                            self.lint_sls('.'.join([parent.replace(os.sep, '.'), sls[:-4]]))
